use std::thread;

use godot::{classes::notify::NodeNotification, prelude::*};
use pipewire::{
    channel::{Receiver, Sender},
    main_loop::MainLoop,
};

#[derive(GodotClass)]
#[class(base=Node)]
/// Node for interacting with Pipewire
/// Intended to not be spawned manually, as the plugin will create a
pub(crate) struct Pipewire {
    base: Base<Node>,

    /// Handle for the thread running the pipewire main loop
    thread_handle: Option<thread::JoinHandle<()>>,
    shutdown_sender: Sender<Terminate>,
}

struct Terminate;

#[godot_api]
impl INode for Pipewire {
    fn init(base: Base<Node>) -> Self {
        let (thread, shutdown_sender) = Self::start_pipewire_thread();
        Self {
            base,
            thread_handle: Some(thread),
            shutdown_sender,
        }
    }
    fn on_notification(&mut self, what: NodeNotification) {
        if what == NodeNotification::PREDELETE
            || what == NodeNotification::UNPARENTED
            || what == NodeNotification::WM_CLOSE_REQUEST
            || what == NodeNotification::EXIT_TREE
        {
            godot_print!("Sending shutdown signal to pipewire thread");
            let _ = self.shutdown_sender.send(Terminate {});
            let handle = self.thread_handle.take().unwrap();
            let _join_result = handle.join();
            godot_print!("Pipewire thread completed");
        }
    }
}

#[godot_api]
impl Pipewire {
    pub const SINGLETON: &'static str = "Pipewire";

    fn start_pipewire_thread() -> (thread::JoinHandle<()>, Sender<Terminate>) {
        godot_print!("Starting pipewire thread");
        let (shutdown_sender, shutdown_receiver) = pipewire::channel::channel::<Terminate>();
        let thread_handle = thread::spawn(|| Self::pipewire_main(shutdown_receiver));
        (thread_handle, shutdown_sender)
    }

    /// Func that actually runs the main loop in the thread
    fn pipewire_main(shutdown_receiver: Receiver<Terminate>) {
        let mainloop = MainLoop::new(None).expect("Failed to create pipewire main loop");

        let _receiver = shutdown_receiver.attach(mainloop.loop_(), {
            godot_print!("Stopping pipewire thread");
            let mainloop = mainloop.clone();
            move |_| mainloop.quit()
        });

        godot_print!("Entering pipewire main loop");
        mainloop.run();
        println!("Pipewire main loop exited");
    }
}
